@@ -542,10 +542,11 @@ int tarobject(struct TarInfo *ti) {
 		stabtmp.st_ino == stab.st_ino)
 	      break;
 	  }
-	  if (conff)
+	  if (conff) {
 	    debug(dbg_eachfiledetail,"tarobject other's obsolete conffile");
 	    /* processarc.c will have copied its hash already. */
 	    continue;
+	  }
 	}
 
         if (does_replace(tc->pkg,&tc->pkg->available,otherpkg)) {
@@ -1187,7 +1188,8 @@ void archivefiles(const char *const *argv) {
   while ((thisarg = *argp++) != NULL) {
     if (setjmp(ejbuf)) {
       error_unwind(ehflag_bombout);
-      if (onerr_abort > 0) break;
+      if (abort_processing)
+        break;
       continue;
     }
     push_error_handler(&ejbuf,print_error_perpackage,thisarg);
