@@ -46,41 +46,6 @@
 static const char* showformat		= "${Package}\t${Version}\n";
 
 
-void ensure_package_clientdata(struct pkginfo *pkg) {
-  if (pkg->clientdata) return;
-  pkg->clientdata= nfmalloc(sizeof(struct perpackagestate));
-  pkg->clientdata->istobe= itb_normal;
-  pkg->clientdata->fileslistvalid= 0;
-  pkg->clientdata->files= 0;
-}
-
-const char *pkgadminfile(struct pkginfo *pkg, const char *whichfile) {
-  static struct varbuf vb;
-  varbufreset(&vb);
-  varbufaddstr(&vb,admindir);
-  varbufaddstr(&vb,"/" INFODIR);
-  varbufaddstr(&vb,pkg->name);
-  varbufaddc(&vb,'.');
-  varbufaddstr(&vb,whichfile);
-  varbufaddc(&vb,0);
-  return vb.buf;
-}
-
-void cu_closepipe(int argc, void **argv) {
-  int *p1= (int*)argv[0];
-  close(p1[0]); close(p1[1]);
-}
-
-void cu_closefile(int argc, void **argv) {
-  FILE *f= (FILE*)(argv[0]);
-  fclose(f);
-}
-
-void cu_closefd(int argc, void **argv) {
-  int ip= *(int*)argv;
-  close(ip);
-}
-
 int pkglistqsortcmp(const void *a, const void *b) {
   const struct pkginfo *pa= *(const struct pkginfo**)a;
   const struct pkginfo *pb= *(const struct pkginfo**)b;
@@ -139,7 +104,7 @@ static void list1package(struct pkginfo *pkg, int *head,
 	if (!pdesc) pdesc= _("(no description available)");
 
 	plen= strlen(pkgl[i]->name);
-	vlen= strlen(versiondescribe(&pkgl[i]->installed.version,vdew_never));
+	vlen = strlen(versiondescribe(&pkgl[i]->installed.version, vdew_nonambig));
 	dlen= strcspn(pdesc, "\n");
 	if (plen > nw) nw = plen;
 	if (vlen > vw) vw = vlen;
@@ -159,7 +124,7 @@ static void list1package(struct pkginfo *pkg, int *head,
   if (!*head) {
     fputs(_("\
 Desired=Unknown/Install/Remove/Purge/Hold\n\
-| Status=Not/Installed/Config-files/Unpacked/Failed-config/Half-installed\n\
+| Status=Not/Inst/Cfg-files/Unpacked/Failed-cfg/Half-inst/trig-aWait/Trig-pend\n\
 |/ Err?=(none)/Hold/Reinst-required/X=both-problems (Status,Err: uppercase=bad)\n"), stdout);
     printf(format,'|','|','/', _("Name"), _("Version"), 40, _("Description"));
     printf("+++-");					/* status */
@@ -173,19 +138,18 @@ Desired=Unknown/Install/Remove/Purge/Hold\n\
   limiteddescription(pkg,dw,&pdesc,&l);
   printf(format,
          "uihrp"[pkg->want],
-         "nUFiHc"[pkg->status],
+         "ncHUFWti"[pkg->status],
          " R?#"[pkg->eflag],
          pkg->name,
-         versiondescribe(&pkg->installed.version,vdew_never),
+         versiondescribe(&pkg->installed.version, vdew_nonambig),
          l, pdesc);
 }
 
 void listpackages(const char *const *argv) {
   struct pkgiterator *it;
   struct pkginfo *pkg;
-  struct pkginfo **pkgl, **pkgf;
-  const char *thisarg;
-  int np, nf, i, head, found;
+  struct pkginfo **pkgl;
+  int np, i, head;
 
   modstatdb_init(admindir,msdbrw_readonly);
 
@@ -199,33 +163,43 @@ void listpackages(const char *const *argv) {
   iterpkgend(it);
   assert(i==np);
 
-  pkgf= m_malloc(sizeof(struct pkginfo*)*np); nf=0;
+  qsort(pkgl, np, sizeof(struct pkginfo*), pkglistqsortcmp);
+  head = 0;
+
   if (!*argv) {
     for (i=0; i<np; i++) {
       pkg= pkgl[i];
       if (pkg->status == stat_notinstalled) continue;
-      pkgf[nf++] = pkg;
+      list1package(pkg, &head, pkgl, np);
     }
   } else {
-    while ((thisarg= *argv++)) {
-      found= 0;
-      for (i=0; i<np; i++) {
-        pkg= pkgl[i];
-        if (fnmatch(thisarg,pkg->name,0)) continue;
-	pkgf[nf++] = pkg; found++;
+    int argc, ip, *found;
+
+    for (argc = 0; argv[argc]; argc++);
+    found = m_malloc(sizeof(int) * argc);
+    memset(found, 0, sizeof(int) * argc);
+
+    for (i = 0; i < np; i++) {
+      pkg = pkgl[i];
+      for (ip = 0; ip < argc; ip++) {
+        if (!fnmatch(argv[ip], pkg->name, 0)) {
+          list1package(pkg, &head, pkgl, np);
+          found[ip]++;
+          break;
+        }
       }
-      if (!found) {
-        fprintf(stderr,_("No packages found matching %s.\n"),thisarg);
-	nerrs++;
+    }
+
+    /* FIXME: we might get non-matching messages for sub-patterns specified
+     * after their super-patterns, due to us skipping on first match. */
+    for (ip = 0; ip < argc; ip++) {
+      if (!found[ip]) {
+        fprintf(stderr, _("No packages found matching %s.\n"), argv[ip]);
+        nerrs++;
       }
     }
   }
 
-  qsort(pkgf,nf,sizeof(struct pkginfo*),pkglistqsortcmp);
-  head=0;
-  for (i=0; i<nf; i++)
-    list1package(pkgf[i],&head,pkgf,nf);
-
   if (ferror(stdout)) werr("stdout");
   if (ferror(stderr)) werr("stderr");  
   modstatdb_shutdown();
@@ -236,18 +210,19 @@ static int searchoutput(struct filenamenode *namenode) {
   struct filepackages *packageslump;
 
   if (namenode->divert) {
-    for (i=0; i<2; i++) {
-      if (namenode->divert->pkg) printf(_("diversion by %s"),namenode->divert->pkg->name);
-      else printf(_("local diversion"));
-      printf(" %s: %s\n", i ? _("to") : _("from"),
-             i ?
-             (namenode->divert->useinstead
-              ? namenode->divert->useinstead->name
-              : namenode->name)
-             :
-             (namenode->divert->camefrom
-              ? namenode->divert->camefrom->name
-              : namenode->name));
+    const char *name_from = namenode->divert->camefrom ?
+                            namenode->divert->camefrom->name : namenode->name;
+    const char *name_to = namenode->divert->useinstead ?
+                          namenode->divert->useinstead->name : namenode->name;
+
+    if (namenode->divert->pkg) {
+      printf(_("diversion by %s from: %s\n"),
+             namenode->divert->pkg->name, name_from);
+      printf(_("diversion by %s to: %s\n"),
+             namenode->divert->pkg->name, name_to);
+    } else {
+      printf(_("local diversion from: %s\n"), name_from);
+      printf(_("local diversion to: %s\n"), name_to);
     }
   }
   found= 0;
@@ -372,10 +347,16 @@ void enqperpackage(const char *const *argv) {
             namenode= file->namenode;
             puts(namenode->name);
             if (namenode->divert && !namenode->divert->camefrom) {
-              if (!namenode->divert->pkg) printf(_("locally diverted"));
-              else if (pkg == namenode->divert->pkg) printf(_("package diverts others"));
-              else printf(_("diverted by %s"),namenode->divert->pkg->name);
-              printf(_(" to: %s\n"),namenode->divert->useinstead->name);
+              if (!namenode->divert->pkg)
+		printf(_("locally diverted to: %s\n"),
+		       namenode->divert->useinstead->name);
+              else if (pkg == namenode->divert->pkg)
+		printf(_("package diverts others to: %s\n"),
+		       namenode->divert->useinstead->name);
+              else
+		printf(_("diverted by %s to: %s\n"),
+		       namenode->divert->pkg->name,
+		       namenode->divert->useinstead->name);
             }
             file= file->next;
           }
@@ -406,8 +387,7 @@ void showpackages(const char *const *argv) {
   struct pkgiterator *it;
   struct pkginfo *pkg;
   struct pkginfo **pkgl;
-  const char *thisarg;
-  int np, i, found;
+  int np, i;
   struct lstitem* fmt = parseformat(showformat);
 
   if (!fmt) {
@@ -436,71 +416,96 @@ void showpackages(const char *const *argv) {
       show1package(fmt,pkg);
     }
   } else {
-    while ((thisarg= *argv++)) {
-      found= 0;
-      for (i=0; i<np; i++) {
-        pkg= pkgl[i];
-        if (fnmatch(thisarg,pkg->name,0)) continue;
-        show1package(fmt,pkg); found++;
+    int argc, ip, *found;
+
+    for (argc = 0; argv[argc]; argc++);
+    found = m_malloc(sizeof(int) * argc);
+    memset(found, 0, sizeof(int) * argc);
+
+    for (i = 0; i < np; i++) {
+      pkg = pkgl[i];
+      for (ip = 0; ip < argc; ip++) {
+        if (!fnmatch(argv[ip], pkg->name, 0)) {
+          show1package(fmt, pkg);
+          found[ip]++;
+          break;
+        }
       }
-      if (!found) {
-        fprintf(stderr,_("No packages found matching %s.\n"),thisarg);
-	nerrs++;
+    }
+
+    /* FIXME: we might get non-matching messages for sub-patterns specified
+     * after their super-patterns, due to us skipping on first match. */
+    for (ip = 0; ip < argc; ip++) {
+      if (!found[ip]) {
+        fprintf(stderr, _("No packages found matching %s.\n"), argv[ip]);
+        nerrs++;
       }
     }
   }
+
   if (ferror(stdout)) werr("stdout");
   if (ferror(stderr)) werr("stderr");  
   freeformat(fmt);
   modstatdb_shutdown();
 }
 
-static void printversion(void) {
-  if (fputs(_("Debian `"), stdout) < 0) werr("stdout");
-  if (fputs(DPKGQUERY, stdout) < 0) werr("stdout");
-  if (fputs(_("' package management program query tool\n"), stdout) < 0)
-    werr("stdout");
-  if (fputs(_( "This is free software; see the GNU General Public Licence version 2 or\n"
-		"later for copying conditions.  There is NO warranty.\n"
-		"See " DPKGQUERY " --licence for copyright and license details.\n"),
-		 stdout) < 0) werr("stdout");
+void
+printversion(void)
+{
+  if (printf(_("Debian `%s' package management program query tool\n"),
+	     DPKGQUERY) < 0) werr("stdout");
+  if (printf(_("This is free software; see the GNU General Public License version 2 or\n"
+	       "later for copying conditions. There is NO warranty.\n"
+	       "See %s --license for copyright and license details.\n"),
+	     DPKGQUERY) < 0) werr("stdout");
 }
 /*
    options that need fixing:
   dpkg --yet-to-unpack                 \n\
   */
-static void usage(void) {
-  if (fprintf (stdout, _("\
-Usage: " DPKGQUERY " [<option>] <command>\n\
-Commands:\n\
-  -s|--status <package-name> ...      display package status details\n\
-  -p|--print-avail <package-name> ... display available version details\n\
-  -L|--listfiles <package-name> ...   list files `owned' by package(s)\n\
-  -l|--list [<pattern> ...]           list packages concisely\n\
-  -W|--show <pattern> ...             show information on package(s)\n\
-  -S|--search <pattern> ...           find package(s) owning file(s)\n\
-  --help | --version                  show this help / version number\n\
-  --licence                           print copyright licensing terms\n\
-\n\
-Options:\n\
-  --admindir=<directory>     Use <directory> instead of %s\n\
-  -f|--showformat=<format>   Use alternative format for --show\n\
-\n\
-Format syntax:\n\
-  A format is a string that will be output for each package. The format\n\
-  can include the standard escape sequences \\n (newline), \\r (carriage\n\
-  return) or \\\\ (plain backslash). Package information can be included\n\
-  by inserting variable references to package fields using the ${var[;width]}\n\
-  syntax. Fields will be right-aligned unless the width is negative in which\n\
-   case left aligenment will be used. \n\
-"),
-	    ADMINDIR) < 0) werr ("stdout");
+void
+usage(void)
+{
+  if (printf(_(
+"Usage: %s [<option> ...] <command>\n"
+"\n"), DPKGQUERY) < 0) werr ("stdout");
+
+  if (printf(_(
+"Commands:\n"
+"  -s|--status <package> ...        Display package status details.\n"
+"  -p|--print-avail <package> ...   Display available version details.\n"
+"  -L|--listfiles <package> ...     List files `owned' by package(s).\n"
+"  -l|--list [<pattern> ...]        List packages concisely.\n"
+"  -W|--show <pattern> ...          Show information on package(s).\n"
+"  -S|--search <pattern> ...        Find package(s) owning file(s).\n"
+"\n")) < 0) werr ("stdout");
+
+  if (printf(_(
+"  -h|--help                        Show this help message.\n"
+"  --version                        Show the version.\n"
+"  --license|--licence              Show the copyright licensing terms.\n"
+"\n")) < 0) werr ("stdout");
+
+  if (printf(_(
+"Options:\n"
+"  --admindir=<directory>           Use <directory> instead of %s.\n"
+"  -f|--showformat=<format>         Use alternative format for --show.\n"
+"\n"), ADMINDIR) < 0) werr ("stdout");
+
+  if (printf(_(
+"Format syntax:\n"
+"  A format is a string that will be output for each package. The format\n"
+"  can include the standard escape sequences \\n (newline), \\r (carriage\n"
+"  return) or \\\\ (plain backslash). Package information can be included\n"
+"  by inserting variable references to package fields using the ${var[;width]}\n"
+"  syntax. Fields will be right-aligned unless the width is negative in which\n"
+"  case left alignment will be used.\n")) < 0) werr ("stdout");
 }
 
 const char thisname[]= "dpkg-query";
 const char printforhelp[]= N_("\
 Use --help for help about querying packages;\n\
-Use --licence for copyright licence and lack of warranty (GNU GPL).\n\
+Use --license for copyright license and lack of warranty (GNU GPL).\n\
 \n");
 
 const struct cmdinfo *cipaction= 0;
@@ -517,18 +522,10 @@ const char *admindir= ADMINDIR;
 const char *instdir= "";
 struct packageinlist *ignoredependss=0;
 
-static void helponly(const struct cmdinfo *cip, const char *value) NONRETURNING;
-static void helponly(const struct cmdinfo *cip, const char *value) {
-  usage(); exit(0);
-}
-static void versiononly(const struct cmdinfo *cip, const char *value) NONRETURNING;
-static void versiononly(const struct cmdinfo *cip, const char *value) {
-  printversion(); exit(0);
-}
-
 static void setaction(const struct cmdinfo *cip, const char *value) {
   if (cipaction)
-    badusage(_("conflicting actions --%s and --%s"),cip->olong,cipaction->olong);
+    badusage(_("conflicting actions -%c (--%s) and -%c (--%s)"),
+             cip->oshort, cip->olong, cipaction->oshort, cipaction->olong);
   cipaction= cip;
 }
 
