@@ -22,11 +22,13 @@
 
 #include "config.h"
 
-#if defined(linux)
+#include <dpkg-def.h>
+
+#if defined(linux) || (defined(__FreeBSD_kernel__) && defined(__GLIBC__))
 #  define OSLinux
 #elif defined(__GNU__)
 #  define OSHURD
-#elif defined(__sparc__)
+#elif defined(__sun)
 #  define OSsunos
 #elif defined(OPENBSD) || defined(__OpenBSD__)
 #  define OSOpenBSD
@@ -43,23 +45,26 @@
 #define MIN_POLL_INTERVAL 20000 /*us*/
 
 #if defined(OSHURD)
-#  include <hurd.h>
-#  include <ps.h>
+#include <hurd.h>
+#include <ps.h>
 #endif
 
 #if defined(OSOpenBSD) || defined(OSFreeBSD) || defined(OSNetBSD)
 #include <sys/param.h>
-#include <sys/user.h>
 #include <sys/proc.h>
 #include <sys/stat.h>
-#include <sys/sysctl.h>
 #include <sys/types.h>
- 
+
 #include <err.h>
-#include <kvm.h>
 #include <limits.h>
 #endif
 
+#ifdef HAVE_KVM_H
+#include <kvm.h>
+#include <sys/sysctl.h>
+#include <sys/user.h>
+#endif
+
 #if defined(OShpux)
 #include <sys/param.h>
 #include <sys/pstat.h>
@@ -86,11 +91,16 @@
 #include <assert.h>
 #include <ctype.h>
 
-#ifdef HAVE_ERROR_H
-#  include <error.h>
+#ifdef HAVE_SYS_CDEFS_H
+#include <sys/cdefs.h>
 #endif
-#ifdef HURD_IHASH_H
-  #include <hurd/ihash.h>
+
+#ifdef HAVE_STDDEF_H
+#include <stddef.h>
+#endif
+
+#ifdef HAVE_ERROR_H
+#include <error.h>
 #endif
 
 static int testmode = 0;
@@ -101,7 +111,6 @@ static int stop = 0;
 static int background = 0;
 static int mpidfile = 0;
 static int signal_nr = 15;
-static const char *signal_str = NULL;
 static int user_id = -1;
 static int runas_uid = -1;
 static int runas_gid = -1;
@@ -115,13 +124,13 @@ static char *execname = NULL;
 static char *startas = NULL;
 static const char *pidfile = NULL;
 static char what_stop[1024];
-static const char *schedule_str = NULL;
 static const char *progname = "";
 static int nicelevel = 0;
+static int umask_value = -1;
 
 static struct stat exec_stat;
 #if defined(OSHURD)
-static struct proc_stat_list *procset;
+static struct proc_stat_list *procset = NULL;
 #endif
 
 
@@ -134,9 +143,13 @@ static struct pid_list *found = NULL;
 static struct pid_list *killed = NULL;
 
 struct schedule_item {
-	enum { sched_timeout, sched_signal, sched_goto, sched_forever } type;
-	int value; /* seconds, signal no., or index into array */
-	/* sched_forever is only seen within parse_schedule and callees */
+	enum {
+		sched_timeout,
+		sched_signal,
+		sched_goto,
+		sched_forever /* Only seen within parse_schedule and callees */
+	} type;
+	int value; /* Seconds, signal no., or index into array */
 };
 
 static int schedule_length;
@@ -151,52 +164,16 @@ static int pid_is_cmd(pid_t pid, const char *name);
 static void check(pid_t pid);
 static void do_pidfile(const char *name);
 static void do_stop(int signal_nr, int quietmode,
-		    int *n_killed, int *n_notkilled, int retry_nr);
+                    int *n_killed, int *n_notkilled, int retry_nr);
 #if defined(OSLinux) || defined(OShpux)
 static int pid_is_exec(pid_t pid, const struct stat *esb);
 #endif
 
 
-#ifdef __GNUC__
 static void fatal(const char *format, ...)
-	NONRETURNPRINTFFORMAT(1, 2);
+	NONRETURNING PRINTFFORMAT(1, 2);
 static void badusage(const char *msg)
 	NONRETURNING;
-#else
-static void fatal(const char *format, ...);
-static void badusage(const char *msg);
-#endif
-
-/* This next part serves only to construct the TVCALC macro, which
- * is used for doing arithmetic on struct timeval's.  It works like this:
- *   TVCALC(result, expression);
- * where result is a struct timeval (and must be an lvalue) and
- * expression is the single expression for both components.  In this
- * expression you can use the special values TVELEM, which when fed a
- * const struct timeval* gives you the relevant component, and
- * TVADJUST.  TVADJUST is necessary when subtracting timevals, to make
- * it easier to renormalise.  Whenver you subtract timeval elements,
- * you must make sure that TVADJUST is added to the result of the
- * subtraction (before any resulting multiplication or what have you).
- * TVELEM must be linear in TVADJUST.
- */
-typedef long tvselector(const struct timeval*);
-static long tvselector_sec(const struct timeval *tv) { return tv->tv_sec; }
-static long tvselector_usec(const struct timeval *tv) { return tv->tv_usec; }
-#define TVCALC_ELEM(result, expr, sec, adj)                           \
-{								      \
-  const long TVADJUST = adj;					      \
-  long (*const TVELEM)(const struct timeval*) = tvselector_##sec;     \
-  (result).tv_##sec = (expr);					      \
-}
-#define TVCALC(result,expr)					      \
-do {								      \
-  TVCALC_ELEM(result, expr, sec, (-1));				      \
-  TVCALC_ELEM(result, expr, usec, (+1000000));			      \
-  (result).tv_sec += (result).tv_usec / 1000000;		      \
-  (result).tv_usec %= 1000000;					      \
-} while(0)
-
 
 static void
 fatal(const char *format, ...)
@@ -207,11 +184,10 @@ fatal(const char *format, ...)
 	va_start(arglist, format);
 	vfprintf(stderr, format, arglist);
 	va_end(arglist);
-	putc('\n', stderr);
+	fprintf(stderr, " (%s)\n", strerror (errno));
 	exit(2);
 }
 
-
 static void *
 xmalloc(int size)
 {
@@ -223,14 +199,62 @@ xmalloc(int size)
 	fatal("malloc(%d) failed", size);
 }
 
-
 static void
 xgettimeofday(struct timeval *tv)
 {
-	if (gettimeofday(tv,0) != 0)
+	if (gettimeofday(tv, NULL) != 0)
 		fatal("gettimeofday failed: %s", strerror(errno));
 }
 
+static void
+tmul(struct timeval *a, int b)
+{
+	a->tv_sec *= b;
+	a->tv_usec *= b;
+	a->tv_sec = a->tv_sec + a->tv_usec / 1000000;
+	a->tv_usec %= 1000000;
+}
+
+static long
+get_open_fd_max(void)
+{
+#ifdef HAVE_GETDTABLESIZE
+	return getdtablesize();
+#else
+	return sysconf(_SC_OPEN_MAX);
+#endif
+}
+
+static void
+daemonize(void)
+{
+	pid_t pid;
+
+	if (quietmode < 0)
+		printf("Detaching to start %s...", startas);
+
+	pid = fork();
+	if (pid < 0)
+		fatal("Unable to do first fork.\n");
+	else if (pid) /* Parent */
+		_exit(0);
+
+	/* Create a new session */
+#ifdef HAVE_SETSID
+	setsid();
+#else
+	setpgid(0, 0);
+#endif
+
+	pid = fork();
+	if (pid < 0)
+		fatal("Unable to do second fork.\n");
+	else if (pid) /* Parent */
+		_exit(0);
+
+	if (quietmode < 0)
+		printf("done.\n");
+}
 
 static void
 push(struct pid_list **list, pid_t pid)
@@ -260,27 +284,31 @@ static void
 do_help(void)
 {
 	printf(
-"start-stop-daemon VERSION for Debian - small and fast C version written by\n"
-"Marek Michalkiewicz <marekm@i17linuxb.ists.pwr.wroc.pl>, public domain.\n"
+"Usage: start-stop-daemon [<option> ...] <command>\n"
 "\n"
-"Usage:\n"
-"  start-stop-daemon -S|--start options ... -- arguments ...\n"
-"  start-stop-daemon -K|--stop options ...\n"
-"  start-stop-daemon -H|--help\n"
-"  start-stop-daemon -V|--version\n"
+"Commands:\n"
+"  -S|--start -- <argument> ...  start a program and pass <arguments> to it\n"
+"  -K|--stop                     stop a program\n"
+"  -H|--help                     print help information\n"
+"  -V|--version                  print version\n"
 "\n"
-"Options (at least one of --exec|--pidfile|--user is required):\n"
-"  -x|--exec <executable>        program to start/check if it is running\n"
+"Matching options (at least one is required):\n"
 "  -p|--pidfile <pid-file>       pid file to check\n"
-"  -c|--chuid <name|uid[:group|gid]>\n"
-"  		change to this user/group before starting process\n"
-"  -u|--user <username>|<uid>    stop processes owned by this user\n"
+"  -x|--exec <executable>        program to start/check if it is running\n"
+"  -n|--name <process-name>      process name to check\n"
+"  -u|--user <username|uid>      process owner to check\n"
+"\n"
+"Options:\n"
 "  -g|--group <group|gid>        run process as this group\n"
-"  -n|--name <process-name>      stop processes with this name\n"
+"  -c|--chuid <name|uid[:group|gid]>\n"
+"                                change to this user/group before starting\n"
+"                                  process\n"
 "  -s|--signal <signal>          signal to send (default TERM)\n"
 "  -a|--startas <pathname>       program to start (default is <executable>)\n"
-"  -C|--chdir <directory>        Change to <directory>(default is /)\n"
+"  -r|--chroot <directory>       chroot to <directory> before starting\n"
+"  -d|--chdir <directory>        change to <directory> (default is /)\n"
 "  -N|--nicelevel <incr>         add incr to the process's nice level\n"
+"  -k|--umask <mask>             change the umask to <mask> before starting\n"
 "  -b|--background               force the process to detach\n"
 "  -m|--make-pidfile             create the pidfile before starting\n"
 "  -R|--retry <schedule>         check whether processes die, and retry\n"
@@ -288,6 +316,7 @@ do_help(void)
 "  -o|--oknodo                   exit status 0 (not 1) if nothing done\n"
 "  -q|--quiet                    be more quiet\n"
 "  -v|--verbose                  be more verbose\n"
+"\n"
 "Retry <schedule> is <item>|/<item>/... where <item> is one of\n"
 " -<signal-num>|[-]<signal-name>  send that signal\n"
 " <timeout>                       wait that many seconds\n"
@@ -298,6 +327,13 @@ do_help(void)
 "              3 = trouble   2 = with --retry, processes wouldn't die\n");
 }
 
+static void
+do_version(void)
+{
+	printf("start-stop-daemon %s for Debian\n\n", VERSION);
+
+	printf("Written by Marek Michalkiewicz, public domain.\n");
+}
 
 static void
 badusage(const char *msg)
@@ -313,7 +349,7 @@ struct sigpair {
 	int signal;
 };
 
-const struct sigpair siglist[] = {
+static const struct sigpair siglist[] = {
 	{ "ABRT",	SIGABRT	},
 	{ "ALRM",	SIGALRM	},
 	{ "FPE",	SIGFPE	},
@@ -335,30 +371,33 @@ const struct sigpair siglist[] = {
 	{ "TTOU",	SIGTTOU	}
 };
 
-static int parse_integer(const char *string, int *value_r) {
+static int
+parse_integer(const char *string, int *value_r)
+{
 	unsigned long ul;
 	char *ep;
 
 	if (!string[0])
 		return -1;
 
-	ul= strtoul(string,&ep,10);
+	ul = strtoul(string, &ep, 10);
 	if (ul > INT_MAX || *ep != '\0')
 		return -1;
 
-	*value_r= ul;
+	*value_r = ul;
 	return 0;
 }
 
-static int parse_signal(const char *signal_str, int *signal_nr)
+static int
+parse_signal(const char *signal_str, int *signal_nr)
 {
 	unsigned int i;
 
 	if (parse_integer(signal_str, signal_nr) == 0)
 		return 0;
 
-	for (i = 0; i < sizeof (siglist) / sizeof (siglist[0]); i++) {
-		if (strcmp (signal_str, siglist[i].name) == 0) {
+	for (i = 0; i < sizeof(siglist) / sizeof(siglist[0]); i++) {
+		if (strcmp(signal_str, siglist[i].name) == 0) {
 			*signal_nr = siglist[i].signal;
 			return 0;
 		}
@@ -366,38 +405,54 @@ static int parse_signal(const char *signal_str, int *signal_nr)
 	return -1;
 }
 
+static int
+parse_umask(const char *string, int *value_r)
+{
+	if (!string[0])
+		return -1;
+
+	errno = 0;
+	*value_r = strtoul(string, NULL, 0);
+	if (errno)
+		return -1;
+	else
+		return 0;
+}
+
 static void
-parse_schedule_item(const char *string, struct schedule_item *item) {
+parse_schedule_item(const char *string, struct schedule_item *item)
+{
 	const char *after_hyph;
 
-	if (!strcmp(string,"forever")) {
+	if (!strcmp(string, "forever")) {
 		item->type = sched_forever;
 	} else if (isdigit(string[0])) {
 		item->type = sched_timeout;
 		if (parse_integer(string, &item->value) != 0)
 			badusage("invalid timeout value in schedule");
 	} else if ((after_hyph = string + (string[0] == '-')) &&
-		   parse_signal(after_hyph, &item->value) == 0) {
+	           parse_signal(after_hyph, &item->value) == 0) {
 		item->type = sched_signal;
 	} else {
 		badusage("invalid schedule item (must be [-]<signal-name>, "
-			 "-<signal-number>, <timeout> or `forever'");
+		         "-<signal-number>, <timeout> or `forever'");
 	}
 }
 
 static void
-parse_schedule(const char *schedule_str) {
+parse_schedule(const char *schedule_str)
+{
 	char item_buf[20];
 	const char *slash;
 	int count, repeatat;
-	ptrdiff_t str_len;
+	size_t str_len;
 
 	count = 0;
 	for (slash = schedule_str; *slash; slash++)
 		if (*slash == '/')
 			count++;
 
-	schedule_length = (count == 0) ? 4 : count+1;
+	schedule_length = (count == 0) ? 4 : count + 1;
 	schedule = xmalloc(sizeof(*schedule) * schedule_length);
 
 	if (count == 0) {
@@ -406,29 +461,29 @@ parse_schedule(const char *schedule_str) {
 		parse_schedule_item(schedule_str, &schedule[1]);
 		if (schedule[1].type != sched_timeout) {
 			badusage ("--retry takes timeout, or schedule list"
-				  " of at least two items");
+			          " of at least two items");
 		}
 		schedule[2].type = sched_signal;
 		schedule[2].value = SIGKILL;
-		schedule[3]= schedule[1];
+		schedule[3] = schedule[1];
 	} else {
 		count = 0;
 		repeatat = -1;
 		while (schedule_str != NULL) {
-			slash = strchr(schedule_str,'/');
-			str_len = slash ? slash - schedule_str : strlen(schedule_str);
-			if (str_len >= (ptrdiff_t)sizeof(item_buf))
+			slash = strchr(schedule_str, '/');
+			str_len = slash ? (size_t)(slash - schedule_str) : strlen(schedule_str);
+			if (str_len >= sizeof(item_buf))
 				badusage("invalid schedule item: far too long"
-					 " (you must delimit items with slashes)");
+				         " (you must delimit items with slashes)");
 			memcpy(item_buf, schedule_str, str_len);
 			item_buf[str_len] = 0;
-			schedule_str = slash ? slash+1 : NULL;
+			schedule_str = slash ? slash + 1 : NULL;
 
 			parse_schedule_item(item_buf, &schedule[count]);
 			if (schedule[count].type == sched_forever) {
 				if (repeatat >= 0)
 					badusage("invalid schedule: `forever'"
-						 " appears more than once");
+					         " appears more than once");
 				repeatat = count;
 				continue;
 			}
@@ -465,17 +520,22 @@ parse_options(int argc, char * const *argv)
 		{ "exec",	  1, NULL, 'x'},
 		{ "chuid",	  1, NULL, 'c'},
 		{ "nicelevel",	  1, NULL, 'N'},
-		{ "background",   0, NULL, 'b'},
+		{ "umask",	  1, NULL, 'k'},
+		{ "background",	  0, NULL, 'b'},
 		{ "make-pidfile", 0, NULL, 'm'},
- 		{ "retry",        1, NULL, 'R'},
-		{ "chdir",        1, NULL, 'd'},
-		{ NULL,		0, NULL, 0}
+		{ "retry",	  1, NULL, 'R'},
+		{ "chdir",	  1, NULL, 'd'},
+		{ NULL,		  0, NULL, 0  }
 	};
+	const char *umask_str = NULL;
+	const char *signal_str = NULL;
+	const char *schedule_str = NULL;
 	int c;
 
 	for (;;) {
-		c = getopt_long(argc, argv, "HKSVa:n:op:qr:s:tu:vx:c:N:bmR:g:d:",
-				longopts, (int *) 0);
+		c = getopt_long(argc, argv,
+		                "HKSVa:n:op:qr:s:tu:vx:c:N:k:bmR:g:d:",
+		                longopts, NULL);
 		if (c == -1)
 			break;
 		switch (c) {
@@ -489,7 +549,7 @@ parse_options(int argc, char * const *argv)
 			start = 1;
 			break;
 		case 'V':  /* --version */
-			printf("start-stop-daemon " VERSION "\n");
+			do_version();
 			exit(0);
 		case 'a':  /* --startas <pathname> */
 			startas = optarg;
@@ -537,6 +597,9 @@ parse_options(int argc, char * const *argv)
 		case 'N':  /* --nice */
 			nicelevel = atoi(optarg);
 			break;
+		case 'k':  /* --umask <mask> */
+			umask_str = optarg;
+			break;
 		case 'b':  /* --background */
 			background = 1;
 			break;
@@ -555,7 +618,7 @@ parse_options(int argc, char * const *argv)
 	}
 
 	if (signal_str != NULL) {
-		if (parse_signal (signal_str, &signal_nr) != 0)
+		if (parse_signal(signal_str, &signal_nr) != 0)
 			badusage("signal value must be numeric or name"
 				 " of signal (KILL, INT, ...)");
 	}
@@ -564,6 +627,11 @@ parse_options(int argc, char * const *argv)
 		parse_schedule(schedule_str);
 	}
 
+	if (umask_str != NULL) {
+		if (parse_umask(umask_str, &umask_value) != 0)
+			badusage("umask value must be a positive number");
+	}
+
 	if (start == stop)
 		badusage("need one of --start or --stop");
 
@@ -577,7 +645,7 @@ parse_options(int argc, char * const *argv)
 		badusage("--start needs --exec or --startas");
 
 	if (mpidfile && pidfile == NULL)
-		badusage("--make-pidfile is only relevant with --pidfile");
+		badusage("--make-pidfile requires --pidfile");
 
 	if (background && !start)
 		badusage("--background is only relevant with --start");
@@ -588,16 +656,27 @@ parse_options(int argc, char * const *argv)
 static int
 pid_is_exec(pid_t pid, const struct stat *esb)
 {
+	char lname[32];
+	char lcontents[_POSIX_PATH_MAX];
+	const char deleted[] = " (deleted)";
+	int nread;
 	struct stat sb;
-	char buf[32];
 
-	sprintf(buf, "/proc/%d/exe", pid);
-	if (stat(buf, &sb) != 0)
+	sprintf(lname, "/proc/%d/exe", pid);
+	nread = readlink(lname, lcontents, sizeof(lcontents));
+	if (nread == -1)
 		return 0;
+
+	lcontents[nread] = '\0';
+	if (strcmp(lcontents + nread - strlen(deleted), deleted) == 0)
+		lcontents[nread - strlen(deleted)] = '\0';
+
+	if (stat(lcontents, &sb) != 0)
+		return 0;
+
 	return (sb.st_dev == esb->st_dev && sb.st_ino == esb->st_ino);
 }
 
-
 static int
 pid_is_user(pid_t pid, uid_t uid)
 {
@@ -610,7 +689,6 @@ pid_is_user(pid_t pid, uid_t uid)
 	return (sb.st_uid == uid);
 }
 
-
 static int
 pid_is_cmd(pid_t pid, const char *name)
 {
@@ -636,40 +714,71 @@ pid_is_cmd(pid_t pid, const char *name)
 }
 #endif /* OSLinux */
 
-
 #if defined(OSHURD)
+static void
+init_procset(void)
+{
+	struct ps_context *context;
+	error_t err;
+
+	err = ps_context_create(getproc(), &context);
+	if (err)
+		error(1, err, "ps_context_create");
+
+	err = proc_stat_list_create(context, &procset);
+	if (err)
+		error(1, err, "proc_stat_list_create");
+
+	err = proc_stat_list_add_all(procset, 0, 0);
+	if (err)
+		error(1, err, "proc_stat_list_add_all");
+}
+
+static struct proc_stat *
+get_proc_stat (pid_t pid, ps_flags_t flags)
+{
+	struct proc_stat *ps;
+	ps_flags_t wanted_flags = PSTAT_PID | flags;
+
+	if (!procset)
+		init_procset();
+
+	ps = proc_stat_list_pid_proc_stat(procset, pid);
+	if (!ps)
+		return NULL;
+	if (proc_stat_set_flags(ps, wanted_flags))
+		return NULL;
+	if ((proc_stat_flags(ps) & wanted_flags) != wanted_flags)
+		return NULL;
+
+	return ps;
+}
+
 static int
 pid_is_user(pid_t pid, uid_t uid)
 {
-	struct stat sb;
-	char buf[32];
-	struct proc_stat *pstat;
+	struct proc_stat *ps;
 
-	sprintf(buf, "/proc/%d", pid);
-	if (stat(buf, &sb) != 0)
-		return 0;
-	return (sb.st_uid == uid);
-	pstat = proc_stat_list_pid_proc_stat (procset, pid);
-	if (pstat)
-		proc_stat_set_flags (pstat, PSTAT_PID | PSTAT_OWNER_UID);
-	else
-		return 1;
-	return (pstat->owner_uid == uid);
+	ps = get_proc_stat(pid, PSTAT_OWNER_UID);
+	return ps && proc_stat_owner_uid(ps) == uid;
 }
 
 static int
 pid_is_cmd(pid_t pid, const char *name)
 {
-	struct proc_stat *pstat;
-	pstat = proc_stat_list_pid_proc_stat (procset, pid);
-	if (pstat)
-		proc_stat_set_flags (pstat, PSTAT_PID | PSTAT_ARGS);
-	else
-		return 1;
-	return (!strcmp (name, pstat->args));
-}
-#endif /* OSHURD */
+	struct proc_stat *ps;
 
+	ps = get_proc_stat(pid, PSTAT_ARGS);
+	return ps && !strcmp(proc_stat_args(ps), name);
+}
+
+static int
+pid_is_running(pid_t pid)
+{
+	return get_proc_stat(pid, 0) != NULL;
+}
+
+#else /* !OSHURD */
 
 static int
 pid_is_running(pid_t pid)
@@ -679,7 +788,7 @@ pid_is_running(pid_t pid)
 
 	sprintf(buf, "/proc/%d", pid);
 	if (stat(buf, &sb) != 0) {
-		if (errno!=ENOENT)
+		if (errno != ENOENT)
 			fatal("Error stating %s: %s", buf, strerror(errno));
 		return 0;
 	}
@@ -687,16 +796,19 @@ pid_is_running(pid_t pid)
 	return 1;
 }
 
+#endif /* OSHURD */
+
 static void
 check(pid_t pid)
 {
 #if defined(OSLinux) || defined(OShpux)
 	if (execname && !pid_is_exec(pid, &exec_stat))
-#elif defined(OSHURD) || defined(OSFreeBSD) || defined(OSNetBSD)
-    /* I will try this to see if it works */
-	if (execname && !pid_is_cmd(pid, execname))
-#endif
 		return;
+#elif defined(OSHURD) || defined(OSFreeBSD) || defined(OSNetBSD)
+	/* Let's try this to see if it works */
+	if (execname && !pid_is_cmd(pid, execname))
+		return;
+#endif
 	if (userspec && !pid_is_user(pid, user_id))
 		return;
 	if (cmdname && !pid_is_cmd(pid, cmdname))
@@ -710,7 +822,12 @@ static void
 do_pidfile(const char *name)
 {
 	FILE *f;
-	pid_t pid;
+	static pid_t pid = 0;
+
+	if (pid) {
+		check(pid);
+		return;
+	}
 
 	f = fopen(name, "r");
 	if (f) {
@@ -719,13 +836,9 @@ do_pidfile(const char *name)
 		fclose(f);
 	} else if (errno != ENOENT)
 		fatal("open pidfile %s: %s", name, strerror(errno));
-
 }
 
-/* WTA: this  needs to be an autoconf check for /proc/pid existance.
- */
-
-#if defined(OSLinux) || defined (OSsunos) || defined(OSfreebsd)
+#if defined(OSLinux) || defined (OSsunos)
 static void
 do_procinit(void)
 {
@@ -751,95 +864,86 @@ do_procinit(void)
 }
 #endif /* OSLinux */
 
-
 #if defined(OSHURD)
+static int
+check_proc_stat(struct proc_stat *ps)
+{
+	check(ps->pid);
+	return 0;
+}
+
 static void
 do_procinit(void)
 {
-	struct ps_context *context;
-	error_t err;
+	if (!procset)
+		init_procset();
 
-	err = ps_context_create(getproc(), &context);
-	if (err)
-		error(1, err, "ps_context_create");
-
-	err = proc_stat_list_create(context, &procset);
-	if (err)
-		error(1, err, "proc_stat_list_create");
-
-	err = proc_stat_list_add_all(procset, 0, 0);
-	if (err)
-		error(1, err, "proc_stat_list_add_all");
-
-	/* Check all pids */
-	HURD_IHASH_ITERATE (&context->procs, ptr) {
-		struct proc_stat *pstat = ptr;
-		check(pstat->pid);
-	}
+	proc_stat_list_for_each(procset, check_proc_stat);
 }
 #endif /* OSHURD */
 
-
-#if defined(OSOpenBSD) || defined(OSFreeBSD) || defined(OSNetBSD)
+#ifdef HAVE_KVM_H
 static int
 pid_is_cmd(pid_t pid, const char *name)
 {
-        kvm_t *kd;
-        int nentries, argv_len=0;
-        struct kinfo_proc *kp;
-        char  errbuf[_POSIX2_LINE_MAX], buf[_POSIX2_LINE_MAX];
-	char  **pid_argv_p;
-	char  *start_argv_0_p, *end_argv_0_p;
- 
- 
-        kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
-        if (kd == 0)
-                errx(1, "%s", errbuf);
-        if ((kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries)) == 0)
-                errx(1, "%s", kvm_geterr(kd));
-	if ((pid_argv_p = kvm_getargv(kd, kp, argv_len)) == 0)
-		errx(1, "%s", kvm_geterr(kd)); 
+	kvm_t *kd;
+	int nentries, argv_len = 0;
+	struct kinfo_proc *kp;
+	char errbuf[_POSIX2_LINE_MAX], buf[_POSIX2_LINE_MAX];
+	char **pid_argv_p;
+	char *start_argv_0_p, *end_argv_0_p;
+
+	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
+	if (kd == NULL)
+		errx(1, "%s", errbuf);
+	kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries);
+	if (kp == NULL)
+		errx(1, "%s", kvm_geterr(kd));
+	pid_argv_p = kvm_getargv(kd, kp, argv_len);
+	if (pid_argv_p == NULL)
+		errx(1, "%s", kvm_geterr(kd));
 
 	start_argv_0_p = *pid_argv_p;
-	/* find and compare string */
-	  
-	/* find end of argv[0] then copy and cut of str there. */
-	if ((end_argv_0_p = strchr(*pid_argv_p, ' ')) == 0 ) 	
-	/* There seems to be no space, so we have the command
-	 * allready in its desired form. */
-	  start_argv_0_p = *pid_argv_p;
+	/* Find and compare string */
+
+	/* Find end of argv[0] then copy and cut of str there. */
+	end_argv_0_p = strchr(*pid_argv_p, ' ');
+	if (end_argv_0_p == NULL)
+		/* There seems to be no space, so we have the command
+		 * allready in its desired form. */
+		start_argv_0_p = *pid_argv_p;
 	else {
-	  /* Tests indicate that this never happens, since
-	   * kvm_getargv itselfe cuts of tailing stuff. This is
-	   * not what the manpage says, however. */
-	  strncpy(buf, *pid_argv_p, (end_argv_0_p - start_argv_0_p));
-	  buf[(end_argv_0_p - start_argv_0_p) + 1] = '\0';
-	  start_argv_0_p = buf;
+		/* Tests indicate that this never happens, since
+		 * kvm_getargv itself cuts of tailing stuff. This is
+		 * not what the manpage says, however. */
+		strncpy(buf, *pid_argv_p, (end_argv_0_p - start_argv_0_p));
+		buf[(end_argv_0_p - start_argv_0_p) + 1] = '\0';
+		start_argv_0_p = buf;
 	}
-        
+
 	if (strlen(name) != strlen(start_argv_0_p))
-               return 0;
-        return (strcmp(name, start_argv_0_p) == 0) ? 1 : 0;
+		return 0;
+	return (strcmp(name, start_argv_0_p) == 0) ? 1 : 0;
 }
- 
+
 static int
 pid_is_user(pid_t pid, uid_t uid)
 {
 	kvm_t *kd;
-	int nentries;   /* Value not used */
+	int nentries; /* Value not used */
 	uid_t proc_uid;
 	struct kinfo_proc *kp;
-	char  errbuf[_POSIX2_LINE_MAX];
-
+	char errbuf[_POSIX2_LINE_MAX];
 
 	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
-	if (kd == 0)
+	if (kd == NULL)
 		errx(1, "%s", errbuf);
-	if ((kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries)) == 0)
+	kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries);
+	if (kp == NULL)
 		errx(1, "%s", kvm_geterr(kd));
-	if (kp->kp_proc.p_cred )
+	if (kp->kp_proc.p_cred)
 		kvm_read(kd, (u_long)&(kp->kp_proc.p_cred->p_ruid),
-			&proc_uid, sizeof(uid_t));
+		         &proc_uid, sizeof(uid_t));
 	else
 		return 0;
 	return (proc_uid == (uid_t)uid);
@@ -854,9 +958,10 @@ pid_is_exec(pid_t pid, const char *name)
 	char errbuf[_POSIX2_LINE_MAX], *pidexec;
 
 	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf);
-	if (kd == 0)
+	if (kd == NULL)
 		errx(1, "%s", errbuf);
-	if ((kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries)) == 0)
+	kp = kvm_getprocs(kd, KERN_PROC_PID, pid, &nentries);
+	if (kp == NULL)
 		errx(1, "%s", kvm_geterr(kd));
 	pidexec = (&kp->kp_proc)->p_comm;
 	if (strlen(name) != strlen(pidexec))
@@ -864,25 +969,22 @@ pid_is_exec(pid_t pid, const char *name)
 	return (strcmp(name, pidexec) == 0) ? 1 : 0;
 }
 
-
 static void
 do_procinit(void)
 {
 	/* Nothing to do */
 }
-
 #endif /* OSOpenBSD */
 
-
 #if defined(OShpux)
 static int
 pid_is_user(pid_t pid, uid_t uid)
 {
 	struct pst_status pst;
 
-	if (pstat_getproc(&pst, sizeof(pst), (size_t) 0, (int) pid) < 0)
+	if (pstat_getproc(&pst, sizeof(pst), (size_t)0, (int)pid) < 0)
 		return 0;
-	return ((uid_t) pst.pst_uid == uid);
+	return ((uid_t)pst.pst_uid == uid);
 }
 
 static int
@@ -890,7 +992,7 @@ pid_is_cmd(pid_t pid, const char *name)
 {
 	struct pst_status pst;
 
-	if (pstat_getproc(&pst, sizeof(pst), (size_t) 0, (int) pid) < 0)
+	if (pstat_getproc(&pst, sizeof(pst), (size_t)0, (int)pid) < 0)
 		return 0;
 	return (strcmp(pst.pst_ucomm, name) == 0);
 }
@@ -900,10 +1002,10 @@ pid_is_exec(pid_t pid, const struct stat *esb)
 {
 	struct pst_status pst;
 
-	if (pstat_getproc(&pst, sizeof(pst), (size_t) 0, (int) pid) < 0)
+	if (pstat_getproc(&pst, sizeof(pst), (size_t)0, (int)pid) < 0)
 		return 0;
-	return ((dev_t) pst.pst_text.psf_fsid.psfs_id == esb->st_dev
-		&& (ino_t) pst.pst_text.psf_fileid == esb->st_ino);
+	return ((dev_t)pst.pst_text.psf_fsid.psfs_id == esb->st_dev
+		&& (ino_t)pst.pst_text.psf_fileid == esb->st_ino);
 }
 
 static void
@@ -914,19 +1016,18 @@ do_procinit(void)
 	int idx = 0;
 
 	while ((count = pstat_getproc(pst, sizeof(pst[0]), 10, idx)) > 0) {
-                for (i = 0; i < count; i++)
+		for (i = 0; i < count; i++)
 			check(pst[i].pst_pid);
-                idx = pst[count - 1].pst_idx + 1;
+		idx = pst[count - 1].pst_idx + 1;
 	}
 }
 #endif /* OShpux */
 
-
 static void
 do_findprocs(void)
 {
 	clear(&found);
-	
+
 	if (pidfile)
 		do_pidfile(pidfile);
 	else
@@ -939,47 +1040,48 @@ do_stop(int signal_nr, int quietmode, int *n_killed, int *n_notkilled, int retry
 {
 	struct pid_list *p;
 
- 	do_findprocs();
- 
- 	*n_killed = 0;
- 	*n_notkilled = 0;
- 
- 	if (!found)
- 		return;
- 
- 	clear(&killed);
+	do_findprocs();
+
+	*n_killed = 0;
+	*n_notkilled = 0;
+
+	if (!found)
+		return;
+
+	clear(&killed);
 
 	for (p = found; p; p = p->next) {
 		if (testmode) {
-			printf("Would send signal %d to %d.\n",
-			       signal_nr, p->pid);
+			if (quietmode <= 0)
+				printf("Would send signal %d to %d.\n",
+				       signal_nr, p->pid);
 			(*n_killed)++;
 		} else if (kill(p->pid, signal_nr) == 0) {
 			push(&killed, p->pid);
- 			(*n_killed)++;
+			(*n_killed)++;
 		} else {
-			printf("%s: warning: failed to kill %d: %s\n",
-			       progname, p->pid, strerror(errno));
- 			(*n_notkilled)++;
+			if (signal_nr)
+				printf("%s: warning: failed to kill %d: %s\n",
+				       progname, p->pid, strerror(errno));
+			(*n_notkilled)++;
 		}
 	}
 	if (quietmode < 0 && killed) {
- 		printf("Stopped %s (pid", what_stop);
+		printf("Stopped %s (pid", what_stop);
 		for (p = killed; p; p = p->next)
 			printf(" %d", p->pid);
- 		putchar(')');
- 		if (retry_nr > 0)
- 			printf(", retry #%d", retry_nr);
- 		printf(".\n");
+		putchar(')');
+		if (retry_nr > 0)
+			printf(", retry #%d", retry_nr);
+		printf(".\n");
 	}
 }
 
-
 static void
 set_what_stop(const char *str)
 {
 	strncpy(what_stop, str, sizeof(what_stop));
-	what_stop[sizeof(what_stop)-1] = '\0';
+	what_stop[sizeof(what_stop) - 1] = '\0';
 }
 
 static int
@@ -990,7 +1092,8 @@ run_stop_schedule(void)
 
 	if (testmode) {
 		if (schedule != NULL) {
-			printf("Ignoring --retry in test mode\n");
+			if (quietmode <= 0)
+				printf("Ignoring --retry in test mode\n");
 			schedule = NULL;
 		}
 	}
@@ -1023,11 +1126,9 @@ run_stop_schedule(void)
 		n_notkilled = 0;
 
 		switch (schedule[position].type) {
-
 		case sched_goto:
 			position = value;
 			continue;
-
 		case sched_signal:
 			do_stop(value, quietmode, &n_killed, &n_notkilled, retry_nr++);
 			if (!n_killed)
@@ -1035,7 +1136,6 @@ run_stop_schedule(void)
 			else
 				anykilled = 1;
 			goto next_item;
-
 		case sched_timeout:
  /* We want to keep polling for the processes, to see if they've exited,
   * or until the timeout expires.
@@ -1063,7 +1163,7 @@ run_stop_schedule(void)
 			ratio = 1;
 			for (;;) {
 				xgettimeofday(&before);
-				if (timercmp(&before,&stopat,>))
+				if (timercmp(&before, &stopat, >))
 					goto next_item;
 
 				do_stop(0, 1, &n_killed, &n_notkilled, 0);
@@ -1072,31 +1172,33 @@ run_stop_schedule(void)
 
 				xgettimeofday(&after);
 
-				if (!timercmp(&after,&stopat,<))
+				if (!timercmp(&after, &stopat, <))
 					goto next_item;
 
 				if (ratio < 10)
 					ratio++;
 
- TVCALC(interval,    ratio * (TVELEM(&after) - TVELEM(&before) + TVADJUST));
- TVCALC(maxinterval,          TVELEM(&stopat) - TVELEM(&after) + TVADJUST);
+				timersub(&stopat, &after, &maxinterval);
+				timersub(&after, &before, &interval);
+				tmul(&interval, ratio);
 
-				if (timercmp(&interval,&maxinterval,>))
+				if (interval.tv_sec < 0 || interval.tv_usec < 0)
+					interval.tv_sec = interval.tv_usec = 0;
+
+				if (timercmp(&interval, &maxinterval, >))
 					interval = maxinterval;
 
 				if (interval.tv_sec == 0 &&
 				    interval.tv_usec <= MIN_POLL_INTERVAL)
-				        interval.tv_usec = MIN_POLL_INTERVAL;
+					interval.tv_usec = MIN_POLL_INTERVAL;
 
-				r = select(0,0,0,0,&interval);
+				r = select(0, NULL, NULL, NULL, &interval);
 				if (r < 0 && errno != EINTR)
 					fatal("select() failed for pause: %s",
 					      strerror(errno));
 			}
-
 		default:
 			assert(!"schedule[].type value must be valid");
-
 		}
 
 	next_item:
@@ -1119,12 +1221,12 @@ x_finished:
 	}
 }
 
-
-int main(int argc, char **argv) NONRETURNING;
 int
 main(int argc, char **argv)
 {
 	int devnull_fd = -1;
+	gid_t rgid;
+	uid_t ruid;
 #ifdef HAVE_TIOCNOTTY
 	int tty_fd = -1;
 #endif
@@ -1134,8 +1236,25 @@ main(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
-	if (execname && stat(execname, &exec_stat))
-		fatal("stat %s: %s", execname, strerror(errno));
+	if (execname) {
+		char *fullexecname;
+
+		if (changeroot) {
+			int fullexecname_len = strlen(changeroot) + 1 +
+			                       strlen(execname) + 1;
+
+			fullexecname = xmalloc(fullexecname_len);
+			snprintf(fullexecname, fullexecname_len, "%s/%s",
+			         changeroot, execname);
+		} else
+			fullexecname = execname;
+
+		if (stat(fullexecname, &exec_stat))
+			fatal("stat %s: %s", fullexecname, strerror(errno));
+
+		if (fullexecname != execname)
+			free(fullexecname);
+	}
 
 	if (userspec && sscanf(userspec, "%d", &user_id) != 1) {
 		struct passwd *pw;
@@ -1158,10 +1277,13 @@ main(int argc, char **argv)
 		if (!pw)
 			fatal("user `%s' not found\n", changeuser);
 		runas_uid = pw->pw_uid;
-		if (changegroup == NULL) { /* pass the default group of this user */
-			changegroup = ""; /* just empty */
+		if (changegroup == NULL) {
+			/* Pass the default group of this user */
+			changegroup = ""; /* Just empty */
 			runas_gid = pw->pw_gid;
 		}
+		if (access(pw->pw_dir, F_OK) == 0)
+			setenv("HOME", pw->pw_dir, 1);
 	}
 
 	if (stop) {
@@ -1176,7 +1298,7 @@ main(int argc, char **argv)
 			printf("%s already running.\n", execname ? execname : "process");
 		exit(exitnodo);
 	}
-	if (testmode) {
+	if (testmode && quietmode <= 0) {
 		printf("Would start %s ", startas);
 		while (argc-- > 0)
 			printf("%s ", *argv++);
@@ -1192,25 +1314,14 @@ main(int argc, char **argv)
 		if (nicelevel)
 			printf(", and add %i to the priority", nicelevel);
 		printf(".\n");
-		exit(0);
 	}
+	if (testmode)
+		exit(0);
 	if (quietmode < 0)
 		printf("Starting %s...\n", startas);
 	*--argv = startas;
 	if (background) { /* ok, we need to detach this process */
-		int i;
-		if (quietmode < 0)
-			printf("Detaching to start %s...", startas);
-		i = fork();
-		if (i<0) {
-			fatal("Unable to fork.\n");
-		}
-		if (i) { /* parent */
-			if (quietmode < 0)
-				printf("done.\n");
-			exit(0);
-		}
-		 /* child continues here */
+		daemonize();
 
 #ifdef HAVE_TIOCNOTTY
 		tty_fd=open("/dev/tty", O_RDWR);
@@ -1218,17 +1329,20 @@ main(int argc, char **argv)
 		devnull_fd=open("/dev/null", O_RDWR);
 	}
 	if (nicelevel) {
-		errno=0;
-		if ((nice(nicelevel)==-1) && (errno!=0))
-			fatal("Unable to alter nice level by %i: %s", nicelevel,
-				strerror(errno));
+		errno = 0;
+		if ((nice(nicelevel) == -1) && (errno != 0))
+			fatal("Unable to alter nice level by %i: %s",
+			      nicelevel, strerror(errno));
 	}
-	if (mpidfile && pidfile != NULL) { /* user wants _us_ to make the pidfile :) */
+	if (umask_value >= 0)
+		umask(umask_value);
+	if (mpidfile && pidfile != NULL) {
+		/* User wants _us_ to make the pidfile :) */
 		FILE *pidf = fopen(pidfile, "w");
 		pid_t pidt = getpid();
 		if (pidf == NULL)
-			fatal("Unable to open pidfile `%s' for writing: %s", pidfile,
-				strerror(errno));
+			fatal("Unable to open pidfile `%s' for writing: %s",
+			      pidfile, strerror(errno));
 		fprintf(pidf, "%d\n", pidt);
 		fclose(pidf);
 	}
@@ -1240,39 +1354,45 @@ main(int argc, char **argv)
 	}
 	if (chdir(changedir) < 0)
 		fatal("Unable to chdir() to %s", changedir);
-	if (changeuser != NULL) {
- 		if (setgid(runas_gid))
- 			fatal("Unable to set gid to %d", runas_gid);
-		if (initgroups(changeuser, runas_gid))
-			fatal("Unable to set initgroups() with gid %d", runas_gid);
-		if (setuid(runas_uid))
-			fatal("Unable to set uid to %s", changeuser);
+
+	rgid = getgid();
+	ruid = getuid();
+	if (changegroup != NULL) {
+		if (rgid != (gid_t)runas_gid)
+			if (setgid(runas_gid))
+				fatal("Unable to set gid to %d", runas_gid);
 	}
-	if (background) { /* continue background setup */
+	if (changeuser != NULL) {
+		/* We assume that if our real user and group are the same as
+		 * the ones we should switch to, the supplementary groups
+		 * will be already in place. */
+		if (rgid != (gid_t)runas_gid || ruid != (uid_t)runas_uid)
+			if (initgroups(changeuser, runas_gid))
+				fatal("Unable to set initgroups() with gid %d",
+				      runas_gid);
+
+		if (ruid != (uid_t)runas_uid)
+			if (setuid(runas_uid))
+				fatal("Unable to set uid to %s", changeuser);
+	}
+
+	if (background) {
+		/* Continue background setup */
 		int i;
 #ifdef HAVE_TIOCNOTTY
-		 /* change tty */
+		 /* Change tty */
 		ioctl(tty_fd, TIOCNOTTY, 0);
 		close(tty_fd);
 #endif
-		umask(022); /* set a default for dumb programs */
-		dup2(devnull_fd,0); /* stdin */
-		dup2(devnull_fd,1); /* stdout */
-		dup2(devnull_fd,2); /* stderr */
-#if defined(OShpux)
-		 /* now close all extra fds */
-		for (i=sysconf(_SC_OPEN_MAX)-1; i>=3; --i) close(i);
-#else
-		 /* now close all extra fds */
-		for (i=getdtablesize()-1; i>=3; --i) close(i);
-#endif
+		if (umask_value < 0)
+			umask(022); /* Set a default for dumb programs */
+		dup2(devnull_fd, 0); /* stdin */
+		dup2(devnull_fd, 1); /* stdout */
+		dup2(devnull_fd, 2); /* stderr */
 
-		/* create a new session */
-#ifdef HAVE_SETSID
-		setsid();
-#else
-		setpgid(0,0);
-#endif
+		 /* Now close all extra fds */
+		for (i = get_open_fd_max() - 1; i >= 3; --i)
+			close(i);
 	}
 	execv(startas, argv);
 	fatal("Unable to start %s: %s", startas, strerror(errno));
